/*
 * dither.c - taken form zgv 2.5 and modified for xpcd
 *
 *   Gerd Knorr (kraxel@cs.tu-berlin.de)
 */

/* Zgv v2.5 - GIF, JPEG and PBM/PGM/PPM viewer, for VGA PCs running Linux.
 * Copyright (C) 1993-1995 Russell Marks. See README for license details.
 *
 * readpnm.c - PBM/PGM/PPM loader.
 *             PBM/PGM are loaded as 8-bit, PPM as 24-bit.
 *             if 8-bit display, PPM is dithered.
 *             xv 3:3:2 thumbnails also.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include <X11/Intrinsic.h>  /* needed for xpcd.h */
#include "xpcd.h"

void ditherline(unsigned char *theline, /* rgb-data      */
		unsigned long *dest,    /* dithered data */
		int linenum,            /* line          */
		int width,              /* line width    */
		int *evenerr,           /* need for ...  */
		int *odderr)            /* ... dithering */
{
  int x,y,lx;
  int c0,c1,c2,times2;
  int terr0,terr1,terr2,actual0,actual1,actual2;
  int start,addon,r,g,b;
  int *thiserr;
  int *nexterr;
  
  y=linenum;
  if((y&1)==0) {
    start=0; addon=1;
    thiserr=evenerr+3; nexterr=odderr+width*3;
  } else {
    start=width-1; addon=-1;
    thiserr=odderr+3; nexterr=evenerr+width*3;
  }
  nexterr[0]=nexterr[1]=nexterr[2]=0;
  x=start;
  
  for(lx=0;lx<width;lx++) {
    b=theline[x*3] >> 1;
    g=theline[x*3+1] >> 1;
    r=theline[x*3+2] >> 1;
    
    terr0=r+((thiserr[0]+8)>>4);
    terr1=g+((thiserr[1]+8)>>4);
    terr2=b+((thiserr[2]+8)>>4);

    /* is this going to screw up on white? */  

    actual0=(terr0>>5)*127/3;
    actual1=(terr1>>4)*127/7;
    actual2=(terr2>>5)*127/3;

    if(actual0<0) actual0=0; if(actual0>127) actual0=127;
    if(actual1<0) actual1=0; if(actual1>127) actual1=127;
    if(actual2<0) actual2=0; if(actual2>127) actual2=127;
    
    c0=terr0-actual0;
    c1=terr1-actual1;
    c2=terr2-actual2;
    
    times2=(c0<<1);
    nexterr[-3] =c0; c0+=times2;
    nexterr[ 3]+=c0; c0+=times2;
    nexterr[ 0]+=c0; c0+=times2;
    thiserr[ 3]+=c0;
    
    times2=(c1<<1);
    nexterr[-2] =c1; c1+=times2;
    nexterr[ 4]+=c1; c1+=times2;
    nexterr[ 1]+=c1; c1+=times2;
    thiserr[ 4]+=c1;
    
    times2=(c2<<1);
    nexterr[-1] =c2; c2+=times2;
    nexterr[ 5]+=c2; c2+=times2;
    nexterr[ 2]+=c2; c2+=times2;
    thiserr[ 5]+=c2;

    dest[x]=colors[(actual0>>5)*32+(actual1>>4)*4+(actual2>>5)];

    thiserr+=3;
    nexterr-=3;
    x+=addon;
  }
}
