/*
 * image.c - (c) 1995 Gerd Knorr (kraxel@cs.tu-berlin.de)
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "config.h"

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xlib.h>
#include <X11/Shell.h>
#include <X11/Xaw/XawInit.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Simple.h>

#include "xpcd.h"

/*------------------------------------------------------------------------*/

void ImageDestroyCB(Widget widget, XtPointer client_data, XtPointer calldata)
{
  struct SHELL_INFO     *sinfo;
  Display               *display;

  sinfo = (struct SHELL_INFO*)client_data;
  display = XtDisplay(app_shell);
  XFreePixmap(display,sinfo->pixmap);
  XFreeGC(display,sinfo->gc);
  XFreeGC(display,sinfo->wgc);
  if (sinfo->workproc) {
    XtRemoveWorkProc(sinfo->workproc);
  }
  if (sinfo->fp) {
    fclose(sinfo->fp);
  }
  if (sinfo->ximage) {
    XDestroyImage(sinfo->ximage);
  }
  free(sinfo);
}

/*------------------------------------------------------------------------*/

void ImageRunIt(struct SHELL_INFO *sinfo, int size)
{
  char           call[512];
  int            rx1,rx2,ry1,ry2,scale;

  scale = (1 << size);
  if ((sinfo->x2 == 0) || (sinfo->y2 == 0)) {
    sprintf(call,"hpcdtoppm %s -%i %s",
            parse_options,
            size,
            sinfo->filename); }
  else {
    switch (sinfo->orientation) {
    case 0:
      rx1=sinfo->x1;
      rx2=sinfo->x2;
      ry1=sinfo->y1;
      ry2=sinfo->y2;
      break;
    case 1:
      rx1=sinfo->y - sinfo->y2 - 1;
      rx2=sinfo->y - sinfo->y1 - 1;
      ry1=sinfo->x1;
      ry2=sinfo->x2;
      break;
    case 2:
      rx1=sinfo->x - sinfo->x2 - 1;
      rx2=sinfo->x - sinfo->x1 - 1;
      ry1=sinfo->y - sinfo->y2 - 1;
      ry2=sinfo->y - sinfo->y1 - 1;
      break;
    case 3:
      rx1=sinfo->y1;
      rx2=sinfo->y2;
      ry1=sinfo->x - sinfo->x2 - 1;
      ry2=sinfo->x - sinfo->x1 - 1;
      break;
    default:
      fprintf(stderr,"internal error\n");
      exit(1);
    }
    sprintf(call,"hpcdtoppm %s -%i -S %i-%i %i-%i %s",
            parse_options,
            size,
            rx1*scale/4,rx2*scale/4,
            ry1*scale/4,ry2*scale/4,
            sinfo->filename); }
  run(call);
}

/*------------------------------------------------------------------------*/

void Size1StubCB(Widget widget, XtPointer client_data, XtPointer calldata)
{
  ImageRunIt((struct SHELL_INFO*) client_data,1);
}  

void Size2StubCB(Widget widget, XtPointer client_data, XtPointer calldata)
{
  ImageRunIt((struct SHELL_INFO*) client_data,2);
}  

void Size3StubCB(Widget widget, XtPointer client_data, XtPointer calldata)
{
  ImageRunIt((struct SHELL_INFO*) client_data,3);
}  

void Size4StubCB(Widget widget, XtPointer client_data, XtPointer calldata)
{
  ImageRunIt((struct SHELL_INFO*) client_data,4);
}  

void Size5StubCB(Widget widget, XtPointer client_data, XtPointer calldata)
{
  ImageRunIt((struct SHELL_INFO*) client_data,5);
}

/*------------------------------------------------------------------------*/

void ImageExposeEvent(Widget widget,
		      XtPointer client_data,
		      XExposeEvent* event)
{
  struct SHELL_INFO *sinfo;
  XGCValues         values;

  sinfo = (struct SHELL_INFO*)client_data;
  if (sinfo->wgc == 0) {
    XtVaGetValues(widget,
		  XtNbackground,&values.foreground,
		  NULL);
    sinfo->wgc = XCreateGC(XtDisplay(widget),
			   XtWindow(widget),
			   GCForeground,&values);
  }

  XCopyArea(XtDisplay(widget),
            sinfo->pixmap,
            XtWindow(widget),
            sinfo->wgc,
            event->x,event->y,
            event->width,event->height,
            event->x,event->y);

   if ((sinfo->x2 != 0) && (sinfo->y2 != 0)) {
    XDrawRectangle(XtDisplay(widget),XtWindow(widget),sinfo->wgc,
                   sinfo->x1, sinfo->y1,
                   sinfo->x2 - sinfo->x1,
                   sinfo->y2 - sinfo->y1); }

}

void ImageMouseEvents(Widget widget, XtPointer client_data, XEvent* event)
{
  static int  startx,starty;
  struct SHELL_INFO *sinfo;
  XExposeEvent      e_event;

  sinfo = (struct SHELL_INFO*)client_data;

  memset(&e_event,0,sizeof(e_event));
  e_event.x = 0;
  e_event.y = 0;
  e_event.width = sinfo->x;
  e_event.height = sinfo->y;

  switch (event->type) {
  case ButtonPress:
    if (event->xbutton.button == Button1) {
      startx = event->xbutton.x;
      starty = event->xbutton.y; }
    break;
  case ButtonRelease:
    if (event->xbutton.button == Button1) {
      if ((event->xbutton.x == startx) || (event->xbutton.y == starty)) {
	sinfo->x1 = 0;
	sinfo->x2 = 0;
	sinfo->y1 = 0;
	sinfo->y2 = 0; }
      else {
	if (event->xbutton.x < startx) {
	  sinfo->x1 = max(0,event->xbutton.x);
	  sinfo->x2 = startx; }
	else {
	  sinfo->x2 = min(sinfo->x-1,event->xbutton.x);
	  sinfo->x1 = startx; }
	if (event->xbutton.y < starty) {
	  sinfo->y1 = max(0,event->xbutton.y);
	  sinfo->y2 = starty; }
	else {
	  sinfo->y2 = min(sinfo->y-1,event->xbutton.y);
	  sinfo->y1 = starty; }}
      ImageExposeEvent(widget,(XtPointer)sinfo,&e_event); }
    break;
  case MotionNotify:
    if (event->xbutton.x < startx) {
      sinfo->x1 = max(0,event->xbutton.x);
      sinfo->x2 = startx; }
    else {
      sinfo->x2 = min(sinfo->x-1,event->xbutton.x);
      sinfo->x1 = startx; }
    if (event->xbutton.y < starty) {
      sinfo->y1 = max(0,event->xbutton.y);
      sinfo->y2 = starty; }
    else {
      sinfo->y2 = min(sinfo->y-1,event->xbutton.y);
      sinfo->y1 = starty; }
    ImageExposeEvent(widget,(XtPointer)sinfo,&e_event);
    break; }
}

/*------------------------------------------------------------------------*/

Boolean LoadImageWP(XtPointer client_data)
{
  struct SHELL_INFO *sinfo;
  int            x;
  int            x1,x2,y1,y2;
  unsigned int   pix;
  unsigned char  *filedata;
  unsigned char  rgbdata[3*384];
  int            i,r,g,b,gray;
  unsigned long  data4ximage[384];
  XExposeEvent   event;
  
  sinfo = (struct SHELL_INFO*)client_data;
  
  /* read data */
  if (!(sinfo->work_count % 2)) {
    fread(sinfo->fgray,768,1,sinfo->fp);
    fread(sinfo->fblue,192,1,sinfo->fp);
    fread(sinfo->fred,192,1,sinfo->fp);
    filedata = sinfo->fgray;
  } else {
    filedata = sinfo->fgray+384;
  }
  /* decode */
  if (use_colors) {
    /* color decodeing */
    for (x = 0, i = 0; x < 384; x++) {
      gray = gray_LUT[filedata[x]];
      b    = gray + blue_LUT[sinfo->fblue[x>>1]];
      r    = gray + red_LUT [sinfo->fred [x>>1]];
      g    = (10*gray - b - 3*r) / 6;
      rgbdata[i++] = gamma_LUT[(b & ~0xff) ? ((b < 0) ? 0 : 255 ) : b];
      rgbdata[i++] = gamma_LUT[(g & ~0xff) ? ((g < 0) ? 0 : 255 ) : g];
      rgbdata[i++] = gamma_LUT[(r & ~0xff) ? ((r < 0) ? 0 : 255 ) : r];
    }
    if (which_visual == VI_PSEUDO_8) {
      /* PseudoColor */
      ditherline(rgbdata,data4ximage,sinfo->work_count,384,
		 sinfo->evenerr,sinfo->odderr);
    } else {
      /* TrueColor */
      for (x = 0, i = 0; x < 384; x++) {
	data4ximage[x]  = TrueColorMap(rgbdata[i+2],rgbdata[i+1],rgbdata[i]);
	i += 3;
      }
    }
  } else {
    /* 16 Grays */
    for (x = 0; x < 384; x++) {
      pix = gray2_LUT[filedata[x]];
      data4ximage[x] = grays[gamma_LUT[pix]/16];
    }
  }
  
  /* fill into ximage, with rotating if nessesary */
  switch(sinfo->orientation) {
  case  0:
    for (x = 0; x < 384; x++) 
      XPutPixel(sinfo->ximage,
		x,sinfo->work_count,data4ximage[x]);
    break;
  case  1:
    for (x = 0; x < 384; x++) 
      XPutPixel(sinfo->ximage,
		sinfo->work_count,sinfo->y-x-1,data4ximage[x]);
    break;
  case  2:
    for (x = 0; x < 384; x++) 
      XPutPixel(sinfo->ximage,
		sinfo->x-x-1,sinfo->y-sinfo->work_count-1,data4ximage[x]);
    break;
  case  3:
    for (x = 0; x < 384; x++) 
      XPutPixel(sinfo->ximage,
		sinfo->x-sinfo->work_count-1,x,data4ximage[x]);
    break;
  default:
    fprintf(stderr,"internal error\n");
    exit(1);
  }
  

  /* copy results to pixmap and screen */
  switch(sinfo->orientation) {
  case  0:
    x1 = 0;
    x2 = 383;
    y1 = y2 = sinfo->work_count;
    break;
  case  1:
    x1 = x2 = sinfo->work_count;
    y1 = 0;
    y2 = 383;
    break;
  case  2:
    x1 = 0;
    x2 = 383;
    y1 = y2 = sinfo->y-sinfo->work_count-1;
    break;
  case  3:
    x1 = x2 = sinfo->x-sinfo->work_count-1;
    y1 = 0;
    y2 = 383;
    break;
  default:
    fprintf(stderr,"internal error\n");
    exit(1);
  }
  
  XPutImage(XtDisplay(app_shell),
            sinfo->pixmap,
            sinfo->gc,
            sinfo->ximage,
            x1,y1,
            x1,y1,
            x2-x1+1,y2-y1+1);
  memset(&event,0,sizeof(event));
  event.x = x1;
  event.y = y1;
  event.width = x2-x1+1;
  event.height = y2-y1+1;

  ImageExposeEvent(sinfo->picture,sinfo,&event);
  
  sinfo->work_count++;
  if (sinfo->work_count < 256) {
    return FALSE;
  } else {
    XDestroyImage(sinfo->ximage);
    sinfo->ximage = NULL;
    fclose(sinfo->fp);
    sinfo->fp = NULL;
    sinfo->workproc = 0;
    return TRUE;
  }
}

/*------------------------------------------------------------------------*/

void ImageButtonCB(Widget widget, XtPointer client_data, XtPointer calldata)
{
  struct COMMAND_INFO   *info;
  struct SHELL_INFO     *sinfo;
  char                  text[512];
  char                  *data;
  int                   size;
  Widget                form,s1,s2,s3,s4,s5,dismiss,picture;
  char                  buffer[1024];

  info = (struct COMMAND_INFO*)client_data;

  if (shells[info->index]) {
    /* exists already */
    XtPopup(shells[info->index],XtGrabNone);
    return;
  }

  sinfo = malloc(sizeof(struct SHELL_INFO));
  if (NULL == sinfo) {
    SetStatus("out of memory");
    return;
  }
  memset(sinfo,0,sizeof(struct SHELL_INFO));
  data = malloc(4*384*256);
  if (NULL == data) {
    SetStatus("out of memory");
    free(sinfo);
    return;
  }
  memset(data,0,4*384*256);
  sinfo->index = info->index;
  strcpy(sinfo->filename,info->filename);

  /* open file */
  sinfo->fp = fopen(sinfo->filename,"r");
  if (sinfo->fp == NULL) {
    sprintf(text,"%s:%s",sinfo->filename,strerror(errno));
    SetStatus(text);
    free(data);
    free(sinfo);
    return;
  }
  fread(buffer,1024,1,sinfo->fp);
  fseek(sinfo->fp,0,SEEK_END);
  size = ftell(sinfo->fp);
  fseek(sinfo->fp,47104,SEEK_SET);

  sinfo->orientation = buffer[0x48] & 3;
  if (sinfo->orientation & 1) {
    sinfo->x = 256;
    sinfo->y = 384;
  } else {
    sinfo->x = 384;
    sinfo->y = 256;
  }

  sinfo->pixmap = XCreatePixmap(XtDisplay(app_shell),
                         RootWindowOfScreen(XtScreen(app_shell)),
                         sinfo->x,sinfo->y,
                         DefaultDepthOfScreen(XtScreen(app_shell)));
  sinfo->gc = XCreateGC(XtDisplay(app_shell),sinfo->pixmap,0,NULL);
  sinfo->ximage = XCreateImage(XtDisplay(app_shell),
			DefaultVisualOfScreen(XtScreen(app_shell)),
			DefaultDepthOfScreen(XtScreen(app_shell)),
			ZPixmap,0,data,
			sinfo->x,sinfo->y,
			8,0);
  XPutImage(XtDisplay(app_shell),
            sinfo->pixmap,
            sinfo->gc,
            sinfo->ximage,
            0,0,
            0,0,
            sinfo->x,sinfo->y);

  /* create Widgets */
  sinfo->shell = shells[info->index] =
    XtVaCreatePopupShell("image", topLevelShellWidgetClass, app_shell,
			 XtNtitle,info->title,
			 NULL);
  form = XtVaCreateManagedWidget("form",formWidgetClass,sinfo->shell,
				 NULL);
  s1 = XtVaCreateManagedWidget("s1",commandWidgetClass,form,
			       XtNleft,XawChainLeft,
			       XtNright,XawChainLeft,
			       XtNtop,XawChainTop,
			       XtNbottom,XawChainTop,
			       NULL);
  XtAddCallback(s1,XtNcallback,
		(XtCallbackProc)Size1StubCB,(XtPointer)sinfo);

  s2 = XtVaCreateManagedWidget("s2",commandWidgetClass,form,
			       XtNleft,XawChainLeft,
			       XtNright,XawChainLeft,
			       XtNtop,XawChainTop,
			       XtNbottom,XawChainTop,
			       XtNfromHoriz,s1,
			       NULL);
  XtAddCallback(s2,XtNcallback,
		(XtCallbackProc)Size2StubCB,(XtPointer)sinfo);

  s3 = XtVaCreateManagedWidget("s3",commandWidgetClass,form,
			       XtNleft,XawChainLeft,
			       XtNright,XawChainLeft,
			       XtNtop,XawChainTop,
			       XtNbottom,XawChainTop,
			       XtNfromHoriz,s2,
			       NULL);
  XtAddCallback(s3,XtNcallback,
		(XtCallbackProc)Size3StubCB,(XtPointer)sinfo);

  if (size > 786432) {
    s4 = XtVaCreateManagedWidget("s4",commandWidgetClass,form,
				 XtNleft,XawChainLeft,
				 XtNright,XawChainLeft,
				 XtNtop,XawChainTop,
				 XtNbottom,XawChainTop,
				 XtNfromHoriz,s3,
				 NULL);
    XtAddCallback(s4,XtNcallback,
		  (XtCallbackProc)Size4StubCB,(XtPointer)sinfo);

    s5 = XtVaCreateManagedWidget("s5",commandWidgetClass,form,
				 XtNleft,XawChainLeft,
				 XtNright,XawChainLeft,
				 XtNtop,XawChainTop,
				 XtNbottom,XawChainTop,
				 XtNfromHoriz,s4,
				 NULL);
    XtAddCallback(s5,XtNcallback,
		  (XtCallbackProc)Size5StubCB,(XtPointer)sinfo);
  }
  dismiss = XtVaCreateManagedWidget("dismiss",commandWidgetClass,form,
				    XtNleft,XawChainLeft,
				    XtNright,XawChainLeft,
				    XtNtop,XawChainTop,
				    XtNbottom,XawChainTop,
				    XtNfromHoriz,(size > 786432) ? s5 : s3,
				    NULL);
  sinfo->picture = picture =
    XtVaCreateManagedWidget("picture",simpleWidgetClass,form,
			    XtNleft,XawChainLeft,
			    XtNright,XawChainLeft,
			    XtNtop,XawChainTop,
			    XtNbottom,XawChainTop,
			    XtNfromVert,s1,
			    XtNheight,sinfo->y,
			    XtNwidth,sinfo->x,
			    NULL);

  XtAddCallback(dismiss,XtNcallback,
		(XtCallbackProc)CancelCB,(XtPointer)sinfo->shell);
  XtAddCallback(sinfo->shell,XtNdestroyCallback,
		(XtCallbackProc)ImageDestroyCB,(XtPointer)sinfo);

  XtAddEventHandler(picture,
                    ExposureMask,
                    False,
                    (XtEventHandler)ImageExposeEvent,
                    (XtPointer)sinfo);
  XtAddEventHandler(picture,
		    ButtonPressMask | ButtonReleaseMask | Button1MotionMask,
		    False,
		    (XtEventHandler)ImageMouseEvents,
		    (XtPointer)sinfo);

  /* display */
  XtPopup(sinfo->shell,XtGrabNone);

  sinfo->workproc = 
    XtAppAddWorkProc(XtWidgetToApplicationContext (app_shell),
		     (XtWorkProc)LoadImageWP, (XtPointer)sinfo);

  return;
}
