/*
 * main.c - (c) 1995 Gerd Knorr (kraxel@cs.tu-berlin.de)
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <math.h>

#include "config.h"

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xlib.h>
#include <X11/Shell.h>
#include <X11/Xaw/XawInit.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Scrollbar.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Viewport.h>

#include "xpcd.h"

/*--- public variables ----------------------------------------------------*/

XtAppContext      app_context;
Widget            app_shell;
Widget            statline,box;

unsigned long     grays[16];
unsigned long     colors[128];
static double     parse_gamma = 1.6;
char              parse_options[512] = "";
char              parse_cdrom[512] = "";
char              parse_viewer[512] = "xv %s";
int               use_colors;
int               load_thumbnails;
int               name_thumbnails;
int               which_visual = 0;

int               wait_for_exit = FALSE;
int               proc_hpcdtoppm = 0;
int               proc_viewer = 0;

int               rgb_red_shift   = 0;
int               rgb_green_shift = 0;
int               rgb_blue_shift  = 0;
int               rgb_red_bits    = 0;
int               rgb_green_bits  = 0;
int               rgb_blue_bits   = 0;

/*--- resources -----------------------------------------------------------*/

static struct RESDATA {
  char     *viewer;
  char     *options;
  char     *cdrom;
  char     *gamma;
  Boolean  usecolors;
  Boolean  loadthumbnails;
  Boolean  namethumbnails;
} resdata;

static XtResource res_desc[] = {
  {
    "viewer",
    XtCString,
    XtRString,
    sizeof(char*),
    XtOffset(struct RESDATA*,viewer),
    XtRString,
    "xv %s"
  },{
    "options",
    XtCString,
    XtRString,
    sizeof(char*),
    XtOffset(struct RESDATA*,options),
    XtRString,
    ""
  },{
    "photocd",
    XtCString,
    XtRString,
    sizeof(char*),
    XtOffset(struct RESDATA*,cdrom),
    XtRString,
    ""
  },{
    "gamma",
    XtCString,
    XtRString,
    sizeof(char*),
    XtOffset(struct RESDATA*,gamma),
    XtRString,
    "1.6"
  },{
    "useColors",
    XtCBoolean,
    XtRBoolean,
    sizeof(Boolean),
    XtOffset(struct RESDATA*,usecolors),
    XtRString,
    "true"
  },{
    "loadThumbnails",
    XtCBoolean,
    XtRBoolean,
    sizeof(Boolean),
    XtOffset(struct RESDATA*,loadthumbnails),
    XtRString,
    "true"
  },{
    "labelThumbnails",
    XtCBoolean,
    XtRBoolean,
    sizeof(Boolean),
    XtOffset(struct RESDATA*,namethumbnails),
    XtRString,
    "true"
  }
};

static XrmOptionDescRec opt_desc[] = {
  { "-viewer",          "viewer",             XrmoptionSepArg, NULL },
  { "-options",         "options",            XrmoptionSepArg, NULL },
  { "-photocd",         "photocd",            XrmoptionSepArg, NULL },
  { "-gamma",           "gamma",              XrmoptionSepArg, NULL },
  { "-usecolors",       "useColors",          XrmoptionSepArg, NULL },
  { "-loadthumbnails",  "loadThumbnails",     XrmoptionSepArg, NULL },
  { "-labelthumbnails", "labelThumbnails",    XrmoptionSepArg, NULL }
};

/*--- fill text to statusline ------------------------------------------*/

void SetStatus(char *text)
{
  XtVaSetValues(statline,
		XtNlabel,text,
		NULL);
}

/*--- popup & -down dialogs --------------------------------------------*/

void PopupCB(Widget widget, XtPointer client_data, XtPointer calldata)
{
  Widget popup;

  popup = (Widget)client_data;
  XtPopup(popup,XtGrabNonexclusive);
}

void CancelCB(Widget widget, XtPointer client_data, XtPointer calldata)
{
  Widget cancel;

  cancel = (Widget)client_data;
  XtPopdown(cancel);
}

/*--- exit ----------------------------------------------------------------*/

void ExitWP(XtPointer client_data)
{
  /* exit if the application is idle,
   * i.e. all the DestroyCallback's are called.
   */
  if (proc_hpcdtoppm || proc_viewer)
    wait_for_exit = TRUE;
  else
    exit(0);
}

void ExitCB(Widget widget, XtPointer client_data, XtPointer calldata)
{
  XtUnmapWidget(app_shell);
  XtDestroyWidget(app_shell);
  XtAppAddWorkProc (XtWidgetToApplicationContext (app_shell),
                    (XtWorkProc)ExitWP, NULL);
}

/*--- Create Widget Tree -------------------------------------------------*/

void CreateWidgets(Widget parent)
{
  Widget  form,viewport;
  Widget  cdrom,files,exit;

  Widget  cd_shell,fi_shell;
  Widget  cd_dialog,fi_dialog;
  
  /* main window */
  form = XtVaCreateManagedWidget("form",formWidgetClass,parent,
				 NULL);

  cdrom= XtVaCreateManagedWidget("cdrom",commandWidgetClass,form,
				 XtNleft,XawChainLeft,
				 XtNright,XawChainLeft,
				 XtNtop,XawChainTop,
				 XtNbottom,XawChainTop,
				 NULL);

  files= XtVaCreateManagedWidget("files",commandWidgetClass,form,
				 XtNleft,XawChainLeft,
				 XtNright,XawChainLeft,
				 XtNtop,XawChainTop,
				 XtNbottom,XawChainTop,
				 XtNfromHoriz,cdrom,
				 NULL);

  exit = XtVaCreateManagedWidget("exit",commandWidgetClass,form,
				 XtNleft,XawChainLeft,
				 XtNright,XawChainLeft,
				 XtNtop,XawChainTop,
				 XtNbottom,XawChainTop,
				 XtNfromHoriz,files,
				 NULL);
  XtAddCallback(exit,XtNcallback,(XtCallbackProc)ExitCB,NULL);
  
  
  statline = XtVaCreateManagedWidget("stat",labelWidgetClass,form,
				     XtNleft,XawChainLeft,
				     XtNright,XawChainRight,
				     XtNtop,XawChainTop,
				     XtNbottom,XawChainTop,
				     XtNfromVert,cdrom,
				     XtNwidth,500,
				     XtNjustify,XtJustifyLeft,
				     NULL);

  viewport = XtVaCreateManagedWidget("box",viewportWidgetClass,form,
				     XtNleft,XawChainLeft,
				     XtNright,XawChainRight,
				     XtNtop,XawChainTop,
				     XtNbottom,XawChainBottom,
				     XtNfromVert,statline,
				     XtNwidth,500,
				     XtNheight,300,
				     XtNallowHoriz,False,
				     XtNallowVert,True,
				     NULL);
  box = XtVaCreateManagedWidget("box",boxWidgetClass,viewport,
				NULL);

  /* cdrom-dialog */
  cd_shell = XtVaCreatePopupShell("cdsh",transientShellWidgetClass,parent,
				  NULL);
  cd_dialog = XtVaCreateManagedWidget("dialog",dialogWidgetClass,cd_shell,
				      XtNvalue,parse_cdrom,
				      NULL);
  XawDialogAddButton(cd_dialog,"ok",
                     (XtCallbackProc)CdromOkCB,(XtPointer)cd_dialog);
  XawDialogAddButton(cd_dialog,"cancel",
		     (XtCallbackProc)CancelCB,(XtPointer)cd_shell);
  XtAddCallback(cdrom,XtNcallback,(XtCallbackProc)PopupCB,(XtPointer)cd_shell);
  
  /* files-dialog */
  fi_shell = XtVaCreatePopupShell("fish",transientShellWidgetClass,parent,
				  NULL);
  fi_dialog = XtVaCreateManagedWidget("dialog",dialogWidgetClass,fi_shell,
				      NULL);
  XawDialogAddButton(fi_dialog,"ok",
                     (XtCallbackProc)FilesOkCB,(XtPointer)fi_dialog);
  XawDialogAddButton(fi_dialog,"cancel",
		     (XtCallbackProc)CancelCB,(XtPointer)fi_shell);
  XtAddCallback(files,XtNcallback,(XtCallbackProc)PopupCB,(XtPointer)fi_shell);
}

/*--- allocate color cells -----------------------------------------------*/

void AllocColors()
{
  XColor       akt_color;
  int          i,mask;
  XVisualInfo  *info, template;
  int          found;
  
  if (use_colors) {
    template.screen   =
      XDefaultScreen(XtDisplay(app_shell));
    template.visualid =
      XVisualIDFromVisual(DefaultVisualOfScreen(XtScreen(app_shell)));
    info = XGetVisualInfo(XtDisplay(app_shell),
			  VisualIDMask | VisualScreenMask,
			  &template, &found);

    if (info->class == TrueColor) {
      which_visual = VI_TRUECOLOR;
      for (i = 0; i < 24; i++) {
	mask = (1 << i);
	if (info->red_mask & mask) rgb_red_bits++;
	else if (!rgb_red_bits) rgb_red_shift++;
	if (info->green_mask & mask) rgb_green_bits++;
	else if (!rgb_green_bits) rgb_green_shift++;
	if (info->blue_mask & mask) rgb_blue_bits++;
	else if (!rgb_blue_bits) rgb_blue_shift++;
      }
      printf("color: bits shift\n");
      printf("red  : %04i %05i\n",rgb_red_bits,rgb_red_shift);
      printf("green: %04i %05i\n",rgb_green_bits,rgb_green_shift);
      printf("blue : %04i %05i\n",rgb_blue_bits,rgb_blue_shift);
    } else if ((info->class == PseudoColor && info->depth == 8)
	       || info->class == StaticColor
	       || info->class == TrueColor) {
      which_visual = VI_PSEUDO_8;
      /* try allocate colors */
      for (i = 0; i < 128; i++) {
	akt_color.red   = ((i>>5)&3)*65535/3;
	akt_color.green = ((i>>2)&7)*65535/7;
	akt_color.blue  = (i&3)*65535/3;
	if (! XAllocColor(XtDisplay(app_shell),
			  DefaultColormapOfScreen(XtScreen(app_shell)),
			  &akt_color)) {
	  /* no success */
	  fprintf(stderr,
		  "Warning: can't allocate 4:8:4 colormap, try using grays\n");
	  use_colors = FALSE;
	  XFreeColors(XtDisplay(app_shell),
		      DefaultColormapOfScreen(XtScreen(app_shell)),
		      colors,i,0);
	  break;
	}
      colors[i] = akt_color.pixel;
      }
    } else {
      fprintf(stderr,"Warning: can't use colors with your visual\n");	    
      use_colors = FALSE;
    }
  }
  if (!use_colors) {
    for (i = 0; i < 16; i++) {
      akt_color.red   = i*65535/15;
      akt_color.green = i*65535/15;
      akt_color.blue  = i*65535/15;
      if (! XAllocColor(XtDisplay(app_shell),
			DefaultColormapOfScreen(XtScreen(app_shell)),
			&akt_color)) {
	fprintf(stderr,"Error: can't allocate 16 grays colormap\n");
	exit(1);
      }
      grays[i] = akt_color.pixel;
    }
  } 
}

long TrueColorMap(int r, int g, int b)
{
  long ret;

  ret  = (r >> (8-rgb_red_bits  )) << rgb_red_shift;
  ret |= (g >> (8-rgb_green_bits)) << rgb_green_shift;
  ret |= (b >> (8-rgb_blue_bits )) << rgb_blue_shift;
  return(ret);
}

/*--- read resources -----------------------------------------------------*/

void Parse()
{
  XtGetApplicationResources(app_shell,&resdata,
			    res_desc,XtNumber(res_desc),
			    NULL,0);

  strcpy(parse_cdrom,resdata.cdrom);
  strcpy(parse_viewer,resdata.viewer);
  strcpy(parse_options,resdata.options);
  parse_gamma = atof(resdata.gamma);
  use_colors = resdata.usecolors;
  load_thumbnails = resdata.loadthumbnails;
  name_thumbnails = resdata.namethumbnails;
}

/*--- load images from command-line --------------------------------------*/

void CheckArgFiles(int argc, char *argv[])
{
  int          i,found;
  char         **namelist;
  char         text[512];
  struct stat  st;
  
  found = 0;
  namelist = NULL;
  for (i = 1; i < argc; i++) {
    if (-1 != stat(argv[i],&st) &&
	S_ISREG(st.st_mode)) {
      if (!(found%10)) {
	namelist = realloc(namelist,(found+10)*sizeof(char*));
      }
      namelist[found] = argv[i];
      found++;
    }
  }

  /* load */
  if (found) {
    sprintf(text,"%i file(s) found\n",found);
    SetStatus(text);
    LoadFileThumbnails(found,namelist);
  }
  
  free(namelist);
}

/*--- main ---------------------------------------------------------------*/

int main(int argc, char *argv[])
{
  app_shell = XtAppInitialize(&app_context,
                              "Xpcd",
                              opt_desc, 7,
                              &argc, argv,
			      NULL,
                              NULL, 0);

  Parse();
  sig_init();
  AllocColors();
  init_gamma_LUT(parse_gamma);
  CreateWidgets(app_shell);

  CheckArgFiles(argc,argv);
  
  XtRealizeWidget(app_shell);
  XtAppMainLoop(app_context);

  /* keep compiler happy */
  return 0;
}
