/*
 * process.c
 *
 * (c) 1994 Gerd Knorr
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>

#include "config.h"

#include <X11/Intrinsic.h>

#include "xpcd.h"

/*------------------------------------------------------------------------*/

char *new_argv[32];             /* hope this is enouth */

/*------------------------------------------------------------------------*/

/* remove terminated child processes from the proc-table */
void sig_child_term (int sig)
{
  int stat = 0;
  
  wait(&stat);
  signal (sig, sig_child_term);
}

void sig_init(void)
{
  signal (SIGCHLD, sig_child_term);
}

/*------------------------------------------------------------------------*/

void InputHpcdtoppm(XtPointer client_data, int *fd, XtInputId *id)
{
  struct PROC_INFO  *info;
  char              buf[513];
  char              cmdline[256];
  int               rc,i,j,w;
  int               fd_pipe[2];

  info = (struct PROC_INFO*)client_data;
  rc = read(info->fd,buf,512);
  if (rc == -1) {
    perror("input_callback");
    return;
  }
  buf[rc] = 0;
  if (rc > 0) {
    /* stderr: print it */
    fprintf(stderr,"%s",buf);
    SetStatus(buf);
  } else {
    /* stderr closed -> hpcdtoppm/viewer done */
    if (info->first) {
      proc_hpcdtoppm--;
      
      /* creating argv for viewer */
      sprintf(cmdline,parse_viewer,info->filename);
      for (i = 0, j = 0, w = FALSE; cmdline[i] != '\0'; i++) {
	if (!w && cmdline[i] != ' ' && cmdline[i] != '\0')
	  new_argv[j++] = &cmdline[i], w = TRUE;
	if (w && (cmdline[i] == ' ' || cmdline[i] == '\0'))
	  w = FALSE;
	if (!w)
	  cmdline[i] = '\0'; }
      new_argv[j] = NULL;
      
      /* running viewer */
      pipe(fd_pipe);
      switch(fork()) {
      case -1:
	perror("xpcd");
	exit(1);
      case 0:
	/* child process */
	signal (SIGCHLD, SIG_DFL);
	setpgid(getpid(),getpid());
	dup2(fd_pipe[1],2);
	close(fd_pipe[0]);
	close(fd_pipe[1]);
	if (-1 == execvp(new_argv[0],new_argv)) {
	  perror(new_argv[0]);
	  exit(1); }
	break;
      default:
	/* parent process */
	dup2(fd_pipe[0],info->fd);
	close(fd_pipe[0]);
	close(fd_pipe[1]);
	info->first = FALSE;
      }
      proc_viewer++;
    } else {
      close(info->fd);
      XtRemoveInput(info->inputproc);
      proc_viewer--;
      unlink(info->filename);
      free(info);
      if (wait_for_exit && !proc_hpcdtoppm && !proc_viewer)
	exit(0);
    }
  }
  return;
}

/*------------------------------------------------------------------------*/

void run(char *options)
{
  static int        count = 0;
  char              filename[32];
  int               fd_pipe[2],file;
  int               i,j,w;
  struct PROC_INFO  *info;

  /* creating argv for hpcdtoppm */
  for (i = 0, j = 0, w = FALSE; options[i] != '\0'; i++) {
    if (!w && options[i] != ' ' && options[i] != '\0')
      new_argv[j++] = &options[i], w = TRUE;
    if (w && (options[i] == ' ' || options[i] == '\0'))
      w = FALSE;
    if (!w)
      options[i] = '\0'; }
  new_argv[j] = NULL;
  
  /* running hpcdtoppm */
  sprintf(filename,"/tmp/xpcd.%i.%i.ppm",getpid(),count++);
  pipe(fd_pipe);
  switch(fork()) {
  case -1:
    perror("xpcd");
    exit(1);
  case 0:
    /* child process */
    signal (SIGCHLD, SIG_DFL);
    if (-1 == (file = creat(filename,0777))) {
      perror(filename);
      exit(1); }
    dup2(file,1);
    dup2(fd_pipe[1],2);
    close(file);
    close(fd_pipe[0]);
    close(fd_pipe[1]);
    if (-1 == execvp(new_argv[0],new_argv)) {
      perror("hpcdtoppm");
      exit(1); }
    break;
  default:
    /* parent process */
    close(fd_pipe[1]);
    info = (struct PROC_INFO*)malloc(sizeof(struct PROC_INFO));
    strcpy(info->filename,filename);
    info->inputproc =
      XtAppAddInput(app_context,fd_pipe[0],(XtPointer)XtInputReadMask,
		    (XtInputCallbackProc)InputHpcdtoppm,(XtPointer)info);
    info->fd = fd_pipe[0];
    info->count = count-1;
    info->first = TRUE;
  }
  proc_hpcdtoppm++;
}
