/*
 * saveimageas.c
 *
 * (c) 1994 Gerd Knorr
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>

#include "config.h"

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xlib.h>
#include <X11/Shell.h>
#include <X11/Xaw/XawInit.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Dialog.h>

#define APPLCLASS   "Saveimageas"

#define BUTTON_PPM    1
#define BUTTON_GIF    2
#define BUTTON_JPEG   3

#define BUTTON_CANCEL 9

/*--- public variables are here ------------------------------------------*/

Widget      app_shell,dialog;
char        **pub_argv;

char        *cat_argv[]      = { "cat",      NULL };
char        *ppmquant_argv[] = { "ppmquant", "-fs",      "256", NULL };
char        *ppmtogif_argv[] = { "ppmtogif", NULL };
char        *cjpeg1_argv[]   = { "cjpeg",    NULL };
char        *cjpeg2_argv[]   = { "cjpeg",    "-quality", NULL,  NULL };

char parse_quality[6]  = "";

/*--- for resource management --------------------------------------------*/

static struct RESDATA {
  char *quality;
} resdata;

static XtResource res_desc[] = {
  {
    "quality",
    XtCString,
    XtRString,
    sizeof(char*),
    XtOffset(struct RESDATA*,quality),
    XtRString,
    ""
  }};

static XrmOptionDescRec opt_desc[] = {
  { "-quality",  "quality",  XrmoptionSepArg, NULL },
};

static String fallback_res[] = {
  "Saveimageas.dialog.value:                    ",
  "Saveimageas.dialog.background:		lightgray",
  
  "Saveimageas.dialog.label:                    enter filename:",
  "Saveimageas.dialog.label.background:         lightgray",
  "Saveimageas.dialog.value.width:              250",
  
  "Saveimageas.dialog.ppm.label:                PPM",
  "Saveimageas.dialog.ppm.background:           gray",
  "Saveimageas.dialog.gif.label:                GIF",
  "Saveimageas.dialog.gif.background:           gray",
  "Saveimageas.dialog.jpeg.label:               JPEG",
  "Saveimageas.dialog.jpeg.background:          gray",
  "Saveimageas.dialog.cancel.label:             Cancel",
  "Saveimageas.dialog.cancel.background:        gray",

  "Saveimageas*beNiceToColormap:                False",

  NULL
};

/*--- local functions ----------------------------------------------------*/

static void ButtonCB(Widget,caddr_t,caddr_t);

static void Parse();

/*--- main function ------------------------------------------------------*/

int main(int argc, char *argv[], char *env[])
{
  XtAppContext      app_context;

  /*----------------------------------------------------------------------*/

  app_shell = XtAppInitialize(&app_context,
			      APPLCLASS,
			      opt_desc, 1,
			      &argc, argv,
			      fallback_res,
			      NULL, 0);

  pub_argv = argv;
  Parse();

  /*----------------------------------------------------------------------*/

  dialog = XtVaCreateManagedWidget("dialog",dialogWidgetClass,app_shell,
				   NULL);

  XawDialogAddButton(dialog,"ppm",
		     (XtCallbackProc)ButtonCB,(caddr_t)BUTTON_PPM);
  XawDialogAddButton(dialog,"gif",
		     (XtCallbackProc)ButtonCB,(caddr_t)BUTTON_GIF);
  XawDialogAddButton(dialog,"jpeg",
		     (XtCallbackProc)ButtonCB,(caddr_t)BUTTON_JPEG);

  XawDialogAddButton(dialog,"cancel",
		     (XtCallbackProc)ButtonCB,(caddr_t)BUTTON_CANCEL);
  
  XtRealizeWidget(app_shell);

  XtAppMainLoop(app_context);

  return(0);
}

/*--- callback's ---------------------------------------------------------*/

static void ButtonCB(Widget widget,caddr_t client_data,caddr_t calldata)
{
  int f_in, f_out, f_pipe[2];
  char *f_oname;

  if (BUTTON_CANCEL == (int)client_data)
    exit(0);

  /* input-file */
  f_in = open(pub_argv[1],O_RDONLY);
  if (-1 == f_in) {
    perror(pub_argv[1]);
    exit(1);
  }

  /* output-file */
  f_oname = XawDialogGetValueString(dialog);
  f_out = creat(f_oname,0666);
  if (-1 == f_out) {
    perror(f_oname);
    exit(1);
  }
  
  switch ((int)client_data) {
  case BUTTON_PPM:
    /* ppm-file - simply copy it with a cat */
    /* io-redirection ... */
    dup2(f_in,0);
    dup2(f_out,1);
    /* ... and call cat */
    execvp(cat_argv[0],cat_argv);
    perror(cat_argv[0]);
    exit(1);
    break;
  case BUTTON_GIF:
    /* gif-file - we need two processes: ppmquant and ppmtogif for that */
    /* create pipe... */
    if (-1 == pipe(f_pipe)) {
      perror("can't create pipe");
      exit(1);
    }
    /* ...fork... */
    switch (fork()) {
    case -1:
      /* error */
      perror("can't fork");
      exit(1);
    case 0:
      /* child process - this will be ppmquant */
      dup2(f_in,0);
      dup2(f_pipe[1],1);
      /* ... and run it */
      execvp(ppmquant_argv[0],ppmquant_argv);
      perror(ppmquant_argv[0]);
      exit(1);
    default:
      /* father process - this will be ppmtogif */
      dup2(f_pipe[0],0);
      dup2(f_out,1);
      /* ... and run it */
      execvp(ppmtogif_argv[0],ppmtogif_argv);
      perror(ppmtogif_argv[0]);
      exit(1);
    }
    break;
  case BUTTON_JPEG:
    /* jpeg-file - compress with cjpeg, optional with quality setting */
    /* io-redirection ... */
    dup2(f_in,0);
    dup2(f_out,1);
    /* ... and call cjpeg */
    if (parse_quality[0] == '\0') {
      /* default quality */
      execvp(cjpeg1_argv[0],cjpeg1_argv);
    } else {
      /* using specifed quality */
      cjpeg2_argv[2] = parse_quality;
      execvp(cjpeg2_argv[0],cjpeg2_argv);
    }
    perror(cjpeg1_argv[0]);
    exit(1);
    break;
  }
}

/*--- other functions ----------------------------------------------------*/

void Parse()
{
  XtGetApplicationResources(app_shell,&resdata,
			    res_desc,XtNumber(res_desc),
			    NULL,0);
  strcpy(parse_quality,resdata.quality);
}
