/*
 * thumbnails.c - (c) 1995 Gerd Knorr (kraxel@cs.tu-berlin.de)
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#include "config.h"

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xlib.h>
#include <X11/Xaw/XawInit.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Command.h>

#include "xpcd.h"

/*------------------------------------------------------------------------*/

Widget           *buttons = NULL;
Widget           *shells  = NULL;
int              imageanz;

/*--- load thumbnails workproc -------------------------------------------*/

Boolean LoadThumbnailsWP(XtPointer client_data)
{
  char                  text[256];
  unsigned char         fgray[384];
  unsigned char         fblue[96];
  unsigned char         fred[96];
  unsigned char         rgbdata[3*96];
  unsigned long         data4ximage[96];
  int                   i,r,g,b,gray;

  int                   x;
  unsigned int          pix;
  struct COMMAND_INFO   *info;

  char                  *data;
  XGCValues             values;
  XTextItem             ti;

  info = (struct COMMAND_INFO*)client_data;

  if (NULL == info->fp) {
    /* open file */
    if (info->overview[0] != '\0') {
      info->fp = fopen(info->overview,"r");
      if (NULL == info->fp) {
	sprintf(text,"%s: %s",info->overview,strerror(errno));
	SetStatus(text);
	return TRUE;
      }
      fseek(info->fp,10240+info->overindex*36864,SEEK_SET);
    } else {
      info->fp = fopen(info->filename,"r");
      if (NULL == info->fp) {
	sprintf(text,"%s: %s",info->filename,strerror(errno));
	SetStatus(text);
	return TRUE;
      }
      fseek(info->fp,8192,SEEK_SET);
    }

    /* create pixmap */
    data = malloc(4*96*64);
    if (NULL == data) {
      SetStatus("out of memory");
      fclose(info->fp);
      return TRUE;
    }
    info->pixmap = XCreatePixmap(XtDisplay(app_shell),
				 RootWindowOfScreen(XtScreen(app_shell)),
				 96,64,
				 DefaultDepthOfScreen(XtScreen(app_shell)));
    XtVaGetValues(info->command,
		  XtNforeground,&values.foreground,
		  NULL);
    info->gc = XCreateGC(XtDisplay(app_shell),info->pixmap,
			 GCForeground,&values);
    info->ximage = XCreateImage(XtDisplay(app_shell),
			DefaultVisualOfScreen(XtScreen(app_shell)),
			DefaultDepthOfScreen(XtScreen(app_shell)),
			ZPixmap,0,data,
			96,64,
			8,0);
    return FALSE;
  }

  /* decode on line per call */
  if (info->work_count < 64) {
    fread(fgray,384,1,info->fp);
    fread(fblue,96,1,info->fp);
    fread(fred,96,1,info->fp);
    if (use_colors) {
      /* color decodeing */
      for (x = 0, i = 0; x < 96; x++) {
	gray = gray_LUT[fgray[x<<1]];
	b    = gray + blue_LUT[fblue[x]];
	r    = gray + red_LUT [fred [x]];
	g    = (10*gray - b - 3*r) / 6;
	rgbdata[i++] = gamma_LUT[(b & ~0xff) ? ((b < 0) ? 0 : 255 ) : b];
	rgbdata[i++] = gamma_LUT[(g & ~0xff) ? ((g < 0) ? 0 : 255 ) : g];
	rgbdata[i++] = gamma_LUT[(r & ~0xff) ? ((r < 0) ? 0 : 255 ) : r];
      }
      if (which_visual == VI_PSEUDO_8) {
	/* PseudoColor */
	ditherline(rgbdata,data4ximage,info->work_count,96,
		   info->evenerr,info->odderr);
	for (x = 0; x < 96; x++) {
	  XPutPixel(info->ximage,x,info->work_count,data4ximage[x]);
	}
      } else {
	/* TrueColor */
	for (x = 0, i = 0; x < 96; x++) {
	  pix = TrueColorMap(rgbdata[i+2],rgbdata[i+1],rgbdata[i]);
	  i += 3;
	  XPutPixel(info->ximage,x,info->work_count,pix);
	}
      }
    } else {
      /* 16 Grays */
      for (x = 0; x < 96; x++) {
	pix = gray2_LUT[fgray[x<<1]];
	XPutPixel(info->ximage,x,info->work_count,grays[gamma_LUT[pix]/16]);
      }
    }
    info->work_count++;
    return FALSE;
  }
    
  XPutImage(XtDisplay(app_shell),
            info->pixmap,
            info->gc,
            info->ximage,
            0,0,
            0,0,
            96,64);
  if (name_thumbnails) {
    ti.chars  = &(info->title[6]);
    ti.nchars = strlen(info->title)-6;
    ti.delta  = 0;
    ti.font   = None;
    XDrawText(XtDisplay(app_shell),info->pixmap,info->gc,
	      7,57,&ti,1);
  }
  
  fclose(info->fp);
  info->fp = NULL;
  XDestroyImage(info->ximage);
  info->ximage = NULL;
  XFreeGC(XtDisplay(app_shell),info->gc);
  info->gc = 0;
  info->workproc = 0;
  info->work_count = 0;

  /* and set it */
  XtVaSetValues(info->command,
		XtNlabel,"",
		XtNbackgroundPixmap,info->pixmap,
		XtNwidth,96,
		XtNheight,64,
		NULL);

  return TRUE;
}

/*--- destroy existing Widgets -------------------------------------------*/

void DeleteButtons()
{
  int i;
  
  if (NULL != buttons) {
    XtUnmanageChildren(buttons,imageanz);
    for (i = 0; i < imageanz; i++) {
      XtDestroyWidget(buttons[i]);
    }
    free(buttons);
    buttons = NULL;
  }
  if (NULL != shells) {
    for (i = 0; i < imageanz; i++) {
      if (shells[i])
	XtDestroyWidget(shells[i]);
    }
    free(shells);
    shells = 0;
  }
  imageanz = 0;
}

/*--- DestroyCallback f. Buttons -----------------------------------------*/

void ButtonDestroyCB(Widget widget, XtPointer client_data, XtPointer calldata)
{
  struct COMMAND_INFO *info;
  Display             *display;

  info = (struct COMMAND_INFO*)client_data;
  if (info->pixmap) {
    display = XtDisplay(app_shell);
    XFreePixmap(display,info->pixmap);
  }
  if (info->workproc) {
    XtRemoveWorkProc(info->workproc);
  }
  if (info->fp) {
    fclose(info->fp);
  }
  if (info->ximage) {
    XDestroyImage(info->ximage);
  }
  if (info->gc) {
    XFreeGC(XtDisplay(app_shell),info->gc);
  }
  free(info);
}

/*--- new PhotoCD --------------------------------------------------------*/

void CdromOkCB(Widget widget, XtPointer client_data, XtPointer calldata)
{
  Widget                dialog,shell;
  String                value;
  char                  text[256],buffer[1024];
  int                   images,i;
  FILE                  *fp;
  struct COMMAND_INFO   *info;
  struct COMMAND_INFO   **infos;

  dialog = (Widget)client_data;
  shell = XtParent(dialog);
  value = XawDialogGetValueString(dialog);
  XtPopdown(shell);

  /* try reading the overview pak */
  sprintf(text,"%s/photo_cd/overview.pcd",value);
  fp = fopen(text,"r");
  if (NULL == fp) {
    sprintf(text,"can't open overview.pcd: %s",strerror(errno));
    SetStatus(text);
    return;
  }
  fread(buffer,1024,1,fp);
  fclose(fp);
  images = buffer[0x0b];
  sprintf(text,"PhotoCD has %i images",images);
  SetStatus(text);

  /* maybe this speeds up fopen for the images ??? */
/*
  sprintf(text,"%s/photo_cd/images",value);
  chdir(text);
*/
  
  /* delete old widgets... */
  DeleteButtons();
  
  /* ... and create new ones */
  infos = malloc(images*sizeof(struct COMMAND_INFO*));
  buttons = malloc(images*sizeof(Widget));
  shells  = malloc(images*sizeof(Widget));
  if (NULL == buttons || NULL == shells || NULL == infos) {
    SetStatus("out of memory");
    return;
  }
  memset(buttons,0,images*sizeof(Widget));
  memset(shells, 0,images*sizeof(Widget));
  imageanz = images;
  for (i = 0; i < imageanz; i++) {
    infos[i] = info = malloc(sizeof(struct COMMAND_INFO));
    memset(info,0,sizeof(struct COMMAND_INFO));
    info->index = i;
    sprintf(info->filename,"%s/photo_cd/images/img%04i.pcd",value,i+1);
    sprintf(info->overview,"%s/photo_cd/overview.pcd",value);
    sprintf(info->title,"Image %i",i+1);
    info->overindex = i;
    sprintf(text,"%i",i+1);
    if (load_thumbnails) {
      buttons[i] = info->command =
	XtVaCreateWidget(text,commandWidgetClass,box,
			 XtNwidth,96,
			 XtNheight,64,
			 NULL);
    } else {
      buttons[i] = info->command =
	XtVaCreateWidget(text,commandWidgetClass,box,
			 NULL);
    }
    XtAddCallback(buttons[i],XtNcallback,
		  (XtCallbackProc)ImageButtonCB,(XtPointer)info);
    XtAddCallback(buttons[i],XtNdestroyCallback,
		  (XtCallbackProc)ButtonDestroyCB,(XtPointer)info);
  }
  if (load_thumbnails) {
    for (i = imageanz-1; i >= 0; i--) {
      infos[i]->workproc =
	XtAppAddWorkProc(XtWidgetToApplicationContext (app_shell),
			 (XtWorkProc)LoadThumbnailsWP, (XtPointer)infos[i]);
    }
  }
  free(infos);
  XtManageChildren(buttons,imageanz);  
}

/*--- new files ----------------------------------------------------------*/

extern int wildmat(const char*, char*);

void FilesOkCB(Widget widget, XtPointer client_data, XtPointer calldata)
{
  Widget                dialog,shell;
  String                value;
  char                  *slash;
  char                  text[512],filemask[32],dirname[256];
  struct stat           st;
  DIR                   *dir;
  struct dirent         *entry;
  int                   found,i;
  char                  **namelist;

  dialog = (Widget)client_data;
  shell = XtParent(dialog);
  value = XawDialogGetValueString(dialog);
  XtPopdown(shell);

  if (0 == strlen(value))
    return;
  slash = strrchr(value,'/');   /* last  '/' */
  if (NULL == slash) {
    if (-1 == stat(value,&st) || !S_ISDIR(st.st_mode)) {
      /* no dir */
      strcpy(dirname,"./");
      strcpy(filemask,value);
    } else {
      /* is a dir */
      strcpy(dirname,value);
      strcpy(filemask,"*.pcd");
    }
  } else if (slash[1] == '\0') {
    /* slash is the last character -> directory */
    if (-1 == stat(value,&st)) {
      sprintf(text,"%s:%s",value,strerror(errno));
      SetStatus(text);
      return;
    }
    if (!S_ISDIR(st.st_mode)) {
      sprintf(text,"%s: not a directory",value);
      SetStatus(text);
      return;
    }
    strcpy(dirname,value);
    strcpy(filemask,"*.pcd");
  } else {
    /* directory ? */
    if (-1 == stat(value,&st) || !S_ISDIR(st.st_mode)) {
      /* no dir */
      slash[1] = '\0';
      strcpy(filemask,slash+1);
      strcpy(dirname,value);
    } else {
      /* is a dir */
      strcpy(filemask,"*.pcd");
      strcpy(dirname,value);
    }
  }

  /* create list of filenames */
  fprintf(stderr,"dir : %s\nmask: %s\n",dirname,filemask);

  dir = opendir(dirname);
  if (NULL == dir) {
    sprintf(text,"%s:%s",dirname,strerror(errno));
    SetStatus(text);
    return;
  }
  found = 0;
  namelist = NULL;
  while (NULL != (entry = readdir(dir))) {
    sprintf(text,"%s/%s",dirname,entry->d_name);
    if (-1 != stat(text,&st) &&
	S_ISREG(st.st_mode) &&
	wildmat(entry->d_name,filemask)) {
      if (!(found%10)) {
	namelist = realloc(namelist,(found+10)*sizeof(char*));
      }
      namelist[found] = malloc(strlen(text)+1);
      strcpy(namelist[found],text);
      found++;
    }
  }
  sprintf(text,"%i file(s) found\n",found);
  SetStatus(text);
  closedir(dir);

  /* load */
  if (found)
    LoadFileThumbnails(found,namelist);

  /* free list memory */
  for (i = 0; i < found; i++) {
    free(namelist[i]);
  }
  free(namelist);
}

void LoadFileThumbnails(int found, char **namelist)
{
  struct COMMAND_INFO   *info;
  struct COMMAND_INFO   **infos;
  char                  *basename;
  int                   i;

  /* delete old widgets... */
  DeleteButtons();
  
  /* ... and create new ones */
  infos = malloc(found*sizeof(struct COMMAND_INFO*));
  buttons = malloc(found*sizeof(Widget));
  shells  = malloc(found*sizeof(Widget));
  if (NULL == buttons || NULL == shells || NULL == infos) {
    SetStatus("out of memory");
    return;
  }
  memset(buttons,0,found*sizeof(Widget));
  memset(shells, 0,found*sizeof(Widget));
  imageanz = found;
  for (i = 0; i < imageanz; i++) {
    infos[i] = info = malloc(sizeof(struct COMMAND_INFO));
    memset(info,0,sizeof(struct COMMAND_INFO));
    info->index = i;
    sprintf(info->filename,"%s",namelist[i]);
    basename=strrchr(namelist[i],'/');
    sprintf(info->title,"Image %s",basename ? basename+1 : namelist[i]);
    buttons[i] = info->command =
      XtVaCreateWidget(basename ? basename+1 : namelist[i],
		       commandWidgetClass,box,
		       XtNwidth,96,
		       XtNheight,64,
		       NULL);
    XtAddCallback(buttons[i],XtNcallback,
		  (XtCallbackProc)ImageButtonCB,(XtPointer)info);
    XtAddCallback(buttons[i],XtNdestroyCallback,
		  (XtCallbackProc)ButtonDestroyCB,(XtPointer)info);
  }
  for (i = imageanz-1; i >= 0; i--) {
    infos[i]->workproc =
      XtAppAddWorkProc(XtWidgetToApplicationContext (app_shell),
		       (XtWorkProc)LoadThumbnailsWP, (XtPointer)infos[i]);
  }
  free(infos);
  XtManageChildren(buttons,imageanz);  
}
